@@ -1,12 +1,14 @@
 use nom::bytes::complete;
 use nom::number::complete::{be_i16, be_i32, be_i64, be_i8, be_u32, be_u8};
 
-use crate::constants::{self,*};
+use tokio::io::{AsyncRead, AsyncReadExt, AsyncWriteExt};
+
+use crate::constants::{self, *};
+use chrono::offset::TimeZone;
+use num_traits::FromPrimitive;
 use std::convert::TryInto;
 use std::fmt;
 use std::path::PathBuf;
-use chrono::offset::TimeZone;
-use num_traits::FromPrimitive;
 
 use super::*;
 use crate::errors::*;
@@ -42,17 +44,22 @@ impl<T> Header<T>
 where
     T: Tag,
 {
-    pub(crate) fn parse<I: std::io::BufRead>(input: &mut I) -> Result<Header<T>, RPMError> {
+    #[cfg(feature = "async-tokio")]
+    pub(crate) async fn parse_async<I: AsyncRead + Unpin>(
+        input: &mut I,
+    ) -> Result<Header<T>, RPMError> {
         let mut buf: [u8; 16] = [0; 16];
-        input.read_exact(&mut buf)?;
+        input.read_exact(&mut buf).await?;
         let index_header = IndexHeader::parse(&buf)?;
         // read rest of header => each index consists of 16 bytes. The index header knows how large the store is.
         let mut buf = vec![0; (index_header.header_size + index_header.num_entries * 16) as usize];
-        input.read_exact(&mut buf)?;
+        input.read_exact(&mut buf).await?;
+        Self::parse_header(index_header, &buf[..])
+    }
 
+    fn parse_header(index_header: IndexHeader, mut bytes: &[u8]) -> Result<Header<T>, RPMError> {
         // parse all entries
         let mut entries: Vec<IndexEntry<T>> = Vec::new();
-        let mut bytes = &buf[..];
         let mut buf_len = bytes.len();
         for _ in 0..index_header.num_entries {
             let (rest, entry) = IndexEntry::parse(bytes)?;
@@ -118,6 +125,28 @@ where
             store,
         })
     }
+    pub(crate) fn parse<I: std::io::BufRead>(input: &mut I) -> Result<Header<T>, RPMError> {
+        let mut buf: [u8; 16] = [0; 16];
+        input.read_exact(&mut buf)?;
+        let index_header = IndexHeader::parse(&buf)?;
+        // read rest of header => each index consists of 16 bytes. The index header knows how large the store is.
+        let mut buf = vec![0; (index_header.header_size + index_header.num_entries * 16) as usize];
+        input.read_exact(&mut buf)?;
+        Self::parse_header(index_header, &buf[..])
+    }
+
+    #[cfg(feature = "async-tokio")]
+    pub(crate) async fn write_async<W: tokio::io::AsyncWrite + Unpin>(
+        &self,
+        out: &mut W,
+    ) -> Result<(), RPMError> {
+        self.index_header.write_async(out).await?;
+        for entry in &self.index_entries {
+            entry.write_index_async(out).await?;
+        }
+        out.write_all(&self.store).await?;
+        Ok(())
+    }
 
     pub(crate) fn write<W: std::io::Write>(&self, out: &mut W) -> Result<(), RPMError> {
         self.index_header.write(out)?;
@@ -159,7 +188,6 @@ where
             })
     }
 
-
     pub(crate) fn get_entry_i16_array_data(&self, tag: T) -> Result<Vec<i16>, RPMError> {
         let entry = self.find_entry_or_err(&tag)?;
         entry
@@ -296,6 +324,21 @@ impl Header<IndexSignatureTag> {
         SignatureHeaderBuilder::<Empty>::new()
     }
 
+    #[cfg(feature = "async-tokio")]
+    pub(crate) async fn parse_signature_async<I: AsyncRead + Unpin>(
+        input: &mut I,
+    ) -> Result<Header<IndexSignatureTag>, RPMError> {
+        let result = Self::parse_async(input).await?;
+
+        let modulo = result.index_header.header_size % 8;
+        if modulo > 0 {
+            let align_size = 8 - modulo;
+            let mut discard = vec![0; align_size as usize];
+            input.read_exact(&mut discard).await?;
+        }
+        Ok(result)
+    }
+
     pub(crate) fn parse_signature<I: std::io::BufRead>(
         input: &mut I,
     ) -> Result<Header<IndexSignatureTag>, RPMError> {
@@ -311,6 +354,20 @@ impl Header<IndexSignatureTag> {
         Ok(result)
     }
 
+    #[cfg(feature = "async-tokio")]
+    pub(crate) async fn write_signature_async<W: tokio::io::AsyncWrite + Unpin>(
+        &self,
+        out: &mut W,
+    ) -> Result<(), RPMError> {
+        self.write_async(out).await?;
+        let modulo = self.index_header.header_size % 8;
+        if modulo > 0 {
+            let expansion = vec![0; 8 - modulo as usize];
+            out.write_all(&expansion).await?;
+        }
+        Ok(())
+    }
+
     pub(crate) fn write_signature<W: std::io::Write>(&self, out: &mut W) -> Result<(), RPMError> {
         self.write(out)?;
         let modulo = self.index_header.header_size % 8;
@@ -414,13 +471,16 @@ impl Header<IndexTag> {
     /// used for headers.
     pub fn get_file_digest_algorithm(&self) -> Result<FileDigestAlgorithm, RPMError> {
         self.get_entry_i32_data(IndexTag::RPMTAG_FILEDIGESTALGO)
-            .and_then(|x| FileDigestAlgorithm::from_i32(x).ok_or_else (|| RPMError::InvalidTagValueEnumVariant {
-                tag: IndexTag::RPMTAG_FILEDIGESTALGO.to_string(),
-                variant: x as u32,
-            }))
+            .and_then(|x| {
+                FileDigestAlgorithm::from_i32(x).ok_or_else(|| {
+                    RPMError::InvalidTagValueEnumVariant {
+                        tag: IndexTag::RPMTAG_FILEDIGESTALGO.to_string(),
+                        variant: x as u32,
+                    }
+                })
+            })
     }
 
-
     /// Extract a the set of contained file names including the additional metadata.
     pub fn get_file_entries(&self) -> Result<Vec<FileEntry>, RPMError> {
         // rpm does not encode it, if it is the default md5
@@ -431,57 +491,72 @@ impl Header<IndexTag> {
         let groups = self.get_entry_string_array_data(IndexTag::RPMTAG_FILEGROUPNAME)?;
         let digests = self.get_entry_string_array_data(IndexTag::RPMTAG_FILEDIGESTS)?;
         let mtimes = self.get_entry_i32_array_data(IndexTag::RPMTAG_FILEMTIMES)?;
-        let sizes = self.get_entry_i64_array_data(IndexTag::RPMTAG_LONGFILESIZES).or_else(|_e| {
-            self.get_entry_i32_array_data(IndexTag::RPMTAG_FILESIZES).map(|file_sizes| {
-                file_sizes.into_iter().map(|file_size| file_size as _ ).collect::<Vec<i64>>()
-            })
-        })?;
+        let sizes = self
+            .get_entry_i64_array_data(IndexTag::RPMTAG_LONGFILESIZES)
+            .or_else(|_e| {
+                self.get_entry_i32_array_data(IndexTag::RPMTAG_FILESIZES)
+                    .map(|file_sizes| {
+                        file_sizes
+                            .into_iter()
+                            .map(|file_size| file_size as _)
+                            .collect::<Vec<i64>>()
+                    })
+            })?;
         let flags = self.get_entry_i32_array_data(IndexTag::RPMTAG_FILEFLAGS)?;
         // @todo
         // let caps = self.get_entry_i32_array_data(IndexTag::RPMTAG_FILECAPS)?;
 
-
         let paths = self.get_file_paths()?;
         let n = paths.len();
 
-        let v = itertools::multizip((paths.into_iter(), users, groups, modes, digests, mtimes, sizes, flags))
-            .try_fold::<Vec<FileEntry>,_,Result<_, RPMError>>(
-                Vec::with_capacity(n),
-                |mut acc, (path,user,group,mode, digest,mtime, size, flags)| {
-                    let digest = if digest.is_empty() { None } else { Some(FileDigest::load_from_str(algorithm, digest)?) };
-                    let utc = chrono::Utc;
-                    acc.push(FileEntry {
-                        path,
-                        ownership: FileOwnership{
-                            user: user.to_owned(),
-                            group: group.to_owned(),
-                        },
-                        mode: FileMode(mode as u16),
-                        modified_at: utc.timestamp( mtime as i64, 0u32),
-                        digest,
-                        category: FileCategory::from_i32(flags).unwrap_or_default(),
-                        size: size as usize,
-                    });
-                    Ok(acc)
-                })?;
+        let v = itertools::multizip((
+            paths.into_iter(),
+            users,
+            groups,
+            modes,
+            digests,
+            mtimes,
+            sizes,
+            flags,
+        ))
+        .try_fold::<Vec<FileEntry>, _, Result<_, RPMError>>(
+            Vec::with_capacity(n),
+            |mut acc, (path, user, group, mode, digest, mtime, size, flags)| {
+                let digest = if digest.is_empty() {
+                    None
+                } else {
+                    Some(FileDigest::load_from_str(algorithm, digest)?)
+                };
+                let utc = chrono::Utc;
+                acc.push(FileEntry {
+                    path,
+                    ownership: FileOwnership {
+                        user: user.to_owned(),
+                        group: group.to_owned(),
+                    },
+                    mode: mode.into(),
+                    modified_at: utc.timestamp(mtime as i64, 0u32),
+                    digest,
+                    category: FileCategory::from_i32(flags).unwrap_or_default(),
+                    size: size as usize,
+                });
+                Ok(acc)
+            },
+        )?;
         Ok(v)
     }
 }
 
-
-/// User facing accessor type representing a file mode
-#[derive(Debug, Clone, Copy, Hash, Eq, PartialEq)]
-pub struct FileMode(pub u16);
-
 /// User facing accessor type representing ownership of a file
 #[derive(Debug, Clone, Hash, Eq, PartialEq)]
-pub struct FileOwnership{ user: String, group: String}
-
-
+pub struct FileOwnership {
+    user: String,
+    group: String,
+}
 
 /// Declaration what category this file belongs to
 /// @todo must be bitflags
-#[derive(Debug,Clone,Copy,Hash,Eq,PartialEq,enum_primitive_derive::Primitive)]
+#[derive(Debug, Clone, Copy, Hash, Eq, PartialEq, enum_primitive_derive::Primitive)]
 #[repr(i32)]
 pub enum FileCategory {
     None = 0i32,
@@ -495,10 +570,8 @@ impl Default for FileCategory {
     }
 }
 
-
-
 #[repr(i32)]
-#[derive(Debug,Clone,Copy,enum_primitive_derive::Primitive)]
+#[derive(Debug, Clone, Copy, enum_primitive_derive::Primitive)]
 pub enum FileDigestAlgorithm {
     // broken and very broken
     Md5 = constants::PGPHASHALGO_MD5,
@@ -535,7 +608,10 @@ pub enum FileDigest {
 }
 
 impl FileDigest {
-    pub fn load_from_str(algorithm: FileDigestAlgorithm, stringly_data: impl AsRef<str>) -> Result<Self, RPMError> {
+    pub fn load_from_str(
+        algorithm: FileDigestAlgorithm,
+        stringly_data: impl AsRef<str>,
+    ) -> Result<Self, RPMError> {
         let hex: Vec<u8> = hex::decode(stringly_data.as_ref())?;
         Ok(match algorithm {
             FileDigestAlgorithm::Md5 if hex.len() == 16 => FileDigest::Md5(hex),
@@ -555,7 +631,7 @@ pub struct FileEntry {
     /// Full path of the file entry and where it will be installed to.
     pub path: PathBuf,
     /// The file mode of the file.
-    pub mode: FileMode,
+    pub mode: types::FileMode,
     /// Defines the owning user and group.
     pub ownership: FileOwnership,
     /// Clocks the last access time.
@@ -701,6 +777,19 @@ impl IndexHeader {
         Ok(())
     }
 
+    #[cfg(feature = "async-tokio")]
+    pub(crate) async fn write_async<W: tokio::io::AsyncWrite + Unpin>(
+        &self,
+        out: &mut W,
+    ) -> Result<(), RPMError> {
+        out.write_all(&self.magic).await?;
+        out.write_all(&self.version.to_be_bytes()).await?;
+        out.write_all(&[0; 4]).await?;
+        out.write_all(&self.num_entries.to_be_bytes()).await?;
+        out.write_all(&self.header_size.to_be_bytes()).await?;
+        Ok(())
+    }
+
     pub(crate) fn new(num_entries: u32, header_size: u32) -> Self {
         IndexHeader {
             magic: HEADER_MAGIC,
@@ -759,6 +848,19 @@ impl<T: num::FromPrimitive + num::ToPrimitive + fmt::Debug + TypeName> IndexEntr
         ))
     }
 
+    #[cfg(feature = "async-tokio")]
+    pub(crate) async fn write_index_async<W: tokio::io::AsyncWrite + Unpin>(
+        &self,
+        out: &mut W,
+    ) -> Result<(), RPMError> {
+        let mut written = out.write(&self.tag.to_u32().unwrap().to_be_bytes()).await?;
+        written += out.write(&self.data.to_u32().to_be_bytes()).await?;
+        written += out.write(&self.offset.to_be_bytes()).await?;
+        written += out.write(&self.num_items.to_be_bytes()).await?;
+        assert_eq!(16, written, "there should be 16 bytes written");
+        Ok(())
+    }
+
     pub(crate) fn write_index<W: std::io::Write>(&self, out: &mut W) -> Result<(), RPMError> {
         let mut written = out.write(&self.tag.to_u32().unwrap().to_be_bytes())?;
         written += out.write(&self.data.to_u32().to_be_bytes())?;
@@ -967,7 +1069,6 @@ impl IndexData {
         }
     }
 
-
     pub(crate) fn as_i32(&self) -> Option<i32> {
         match self {
             IndexData::Int32(s) => {
